/*
 * uhub - A tiny ADC p2p connection hub
 * Copyright (C) 2007, Jan Vidar Krey
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

#include "uhub.h"

#define ACL_CHECK_COMMAND(CMD, STATUS, LIST) \
	if (!strncmp(line, CMD, strlen(CMD)))\
	{\
		data = &line[strlen(CMD)]; \
		data_extra = 0; \
		data[0] = '\0'; \
		data++; \
		\
		data = strip_white_space(data); \
		if (!strlen(data)) \
		{ \
			uhub_log(log_fatal, "ACL parse error on line %d", line_count); \
			return -1; \
		} \
		\
		info = malloc(sizeof(struct access_info)); \
		\
		if (!info) \
		{ \
			uhub_log(log_error, "ACL parse error. Out of memory!"); \
			return -1; \
		} \
		\
		if (strncmp(CMD, "user_", 5) == 0) \
		{ \
			data_extra = strrchr(data, ':'); \
			if (data_extra) \
			{ \
				data_extra[0] = 0; \
				data_extra++; \
			} \
		} \
		\
		memset(info, 0, sizeof(struct access_info)); \
		info->data1 = strdup(data); \
		info->data2 = data_extra ? strdup(data_extra) : 0; \
		info->status = STATUS; \
		list_append(LIST, info); \
		uhub_log(log_debug, "Added ACL entry for '%s' (%s) (line %d, type=%s)", data, data_extra ? data_extra : "", line_count, CMD); \
		return 0; \
	}


static int acl_parse_line(char* line, int line_count, void* ptr_data)
{
	char* pos;
	char* data;
	char* data_extra;
	struct access_info* info = 0;
	struct acl_handle* handle = (struct acl_handle*) ptr_data;
	
	if ((pos = strchr(line, '#')) != NULL)
	{
		pos[0] = 0;
	}
	
	if (strlen(line) == 0)
		return 0;
	
	uhub_log(log_trace, "acl_parse_line(): '%s'", line);
	line = strip_white_space(line);
	
	if (!strlen(line))
	{
		uhub_log(log_fatal, "ACL parse error on line %d", line_count);
		return -1;
	}
	
	uhub_log(log_trace, "acl_parse_line: '%s'", line);
	
	ACL_CHECK_COMMAND("user_reg",   user_auth_regular,   handle->nicks)
	ACL_CHECK_COMMAND("user_op",    user_auth_operator,  handle->nicks)
	ACL_CHECK_COMMAND("user_admin", user_auth_admin,     handle->nicks)
	ACL_CHECK_COMMAND("deny_nick",  user_not_allowed,    handle->nicks)
	ACL_CHECK_COMMAND("deny_ip",    user_not_allowed,    handle->networks)
	ACL_CHECK_COMMAND("ban_nick",   user_banned,         handle->nicks)
	ACL_CHECK_COMMAND("ban_cid",    user_banned,         handle->cids)
	
	uhub_log(log_error, "Unknown ACL command on line %d: '%s'", line_count, line);
	return -1;
}


int acl_initialize(struct uhub_config* config, struct acl_handle* handle)
{
	int ret;
	
	assert(handle);
	memset(handle, 0, sizeof(struct acl_handle));
	
	handle->nicks    = list_create();
	handle->cids     = list_create();
	handle->networks = list_create();
	
	if (!handle->nicks || !handle->cids || !handle->networks)
	{
		uhub_log(log_fatal, "acl_initialize: Out of memory");
		return -1;
	}
	
	ret = file_read_lines(config->file_acl, handle, &acl_parse_line);
	if (ret == -1)
		return -1;
	return 0;
}


int acl_shutdown(struct acl_handle* handle)
{
	if (handle->nicks)
	{
		list_clear(handle->nicks, &free);
		list_destroy(handle->nicks);
		handle->nicks = 0;
	}
	
	if (handle->cids)
	{
		list_clear(handle->cids, &free);
		list_destroy(handle->cids);
		handle->cids = 0;
	}
	
	if (handle->networks)
	{
		list_clear(handle->networks, &free);
		list_destroy(handle->networks);
		handle->networks = 0;
	}
	return 0;
}


struct access_info* acl_get_access_info_by_name(struct acl_handle* handle, const char* name)
{
	struct access_info* info = (struct access_info*) list_get_first(handle->nicks);
	
	while (info)
	{
		if (strcasecmp(info->data1, name) == 0)
		{
			return info;
		}
		info = (struct access_info*) list_get_next(handle->nicks);
	}
	
	return NULL;
}


struct access_info* acl_get_access_info_by_cid(struct acl_handle* handle, const char* cid)
{
	struct access_info* info = (struct access_info*) list_get_first(handle->cids);
	
	while (info)
	{
		if (strcasecmp(info->data1, cid) == 0)
		{
			return info;
		}
		info = (struct access_info*) list_get_next(handle->cids);
	}
	
	return NULL;
}


struct access_info* acl_get_access_info_by_network(struct acl_handle* handle, const char* ip_address)
{
	/* FIXME: implement me! */
	return NULL;
}

/*
 * This will generate the same challenge to the same user, always.
 * The challenge is made up of the time of the user connected
 * seconds since the unix epoch (modulus 1 million)
 * and the SID of the user (0-1 million).
 */
const char* password_generate_challenge(struct user* user)
{
	char buf[32];
	uint64_t tiger_res[3];
	static char tiger_buf[MAX_CID_LEN+1];
	
	snprintf(buf, 32, "%d%d%d", (int) user->tm_connected, (int) user->sid, (int) user->sd);
	
	tiger((uint64_t*) buf, strlen(buf), (uint64_t*) tiger_res);
	base32_encode((unsigned char*) tiger_res, TIGERSIZE, tiger_buf);
	buf[MAX_CID_LEN] = 0;
	
	uhub_log(log_trace, "Generating challenge for user %s: '%s'", user->nick, tiger_buf);
	return (const char*) tiger_buf;
}


int password_verify(struct user* user, const char* password)
{
	char buf[1024];
	struct access_info* access;
	const char* challenge;
	char raw_cid[TIGERSIZE];
	char raw_challenge[TIGERSIZE];
	char password_calc[MAX_CID_LEN+1];
	uint64_t tiger_res[3];
	
	if (!password || !user || strlen(password) != MAX_CID_LEN) return 0;

	access = acl_get_access_info_by_name(user->hub->acl, user->nick);
	if (!access || !access->data2) return 0;

	if (TIGERSIZE+TIGERSIZE+strlen(access->data2) > 1024) return 0;

	challenge = password_generate_challenge(user);

	base32_decode(challenge, (unsigned char*) raw_challenge, MAX_CID_LEN);
	base32_decode(user->cid, (unsigned char*) raw_cid, MAX_CID_LEN);
	
	memcpy(buf, raw_cid, TIGERSIZE);
	memcpy(&buf[TIGERSIZE], (char*) access->data2, strlen(access->data2));
	memcpy(&buf[TIGERSIZE+strlen(access->data2)], raw_challenge, TIGERSIZE);
	
	tiger((uint64_t*) buf, TIGERSIZE+TIGERSIZE+strlen(access->data2), (uint64_t*) tiger_res);
	base32_encode((unsigned char*) tiger_res, TIGERSIZE, password_calc);
	
	uhub_log(log_trace, "Checking password %s against %s", password, password_calc);
	
	if (strcasecmp(password, password_calc) == 0) return 1;
	
	return 0;
}



