/*
 * uhub - A tiny ADC p2p connection hub
 * Copyright (C) 2007, Jan Vidar Krey
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

#include "uhub.h"


static int is_ipv6_supported = -1; /* -1 = CHECK, 0 = NO, 1 = YES */


int net_set_nonblocking(int fd, int toggle)
{
	int ret;
	uhub_log(log_trace, "net_set_nonblocking(): fd=%d", fd);
	ret = ioctl(fd, FIONBIO, &toggle);
    if (ret == -1)
	{
		uhub_log(log_error, "net_set_nonblocking(): ioctl failed (fd=%d): %s", fd, strerror(errno));
		return -1;
	}
	return 0;
}


int net_set_reuseaddress(int fd, int toggle)
{
	int ret;
	uhub_log(log_trace, "net_set_reuseaddress(): fd=%d", fd);
	ret = setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &toggle, sizeof(toggle));
	if (ret == -1)
	{
		uhub_log(log_error, "net_set_reuseaddress(): setsockopt failed (fd=%d): %s", fd, strerror(errno));
		return -1;
	}
	return 0;
}


int net_close(int fd)
{
	return close(fd);
}

int net_accept(int fd)
{
	struct sockaddr_in6 addr;
	socklen_t addr_size;
	int ret = 0;
	addr_size = sizeof(struct sockaddr_in6);
	memset(&addr, 0, addr_size);
	uhub_log(log_trace, "net_accept() fd=%d", fd);
	ret = accept(fd, (struct sockaddr*) &addr, &addr_size);
	
	if (ret == -1)
	{
		uhub_log(log_error, "net_accept(): accept failed (fd=%d, errno=%d, msg=%s)", fd, errno, strerror(errno));
	}
	return ret;
}


int net_is_ipv6_supported()
{
	if (is_ipv6_supported == -1)
	{
		int ret = socket(AF_INET6, SOCK_DGRAM, IPPROTO_UDP);
		if (ret == -1)
		{
			if (errno == EAFNOSUPPORT)
			{
				is_ipv6_supported = 0;
				return 0;
			}
		
		}
		else
		{
			close(ret);
		}
		is_ipv6_supported = 1;
	}
	return is_ipv6_supported;
}


int net_socket_create(int af, int type, int protocol)
{
	int sd = socket(af, type, protocol);
	if (sd == -1)
	{
		uhub_log(log_error, "net_socket_create(): socket failed (errno=%d, msg=%s)", errno, strerror(errno));
	}
	return sd;
}

const char* net_get_peer_address(int fd)
{
	static char address[INET6_ADDRSTRLEN+1];
	struct sockaddr_in6 name;
	socklen_t namelen = sizeof(struct sockaddr_in6);
	int af = net_is_ipv6_supported() ? AF_INET6 : AF_INET;
	
	memset(address, 0, INET6_ADDRSTRLEN);
	
	if (getpeername(fd, (struct sockaddr*) &name, &namelen) != -1)
	{
		inet_ntop(af, (void*) &(&name)->sin6_addr, address, INET6_ADDRSTRLEN);
		if (strncmp(address, "::ffff:", 7) == 0) /* IPv6 mapped IPv4 address. */
		{
			return &address[7];
		}
		uhub_log(log_trace, "net_get_peer_address(): address=%s", address);
		return address;
	}
	else
	{
		uhub_log(log_error, "net_get_peer_address(): getsockname failed (fd=%d, errno=%d, msg=%s)", fd, errno, strerror(errno));
	}
	
	return "0.0.0.0";
}

