/*
 * uhub - A tiny ADC p2p connection hub
 * Copyright (C) 2007, Jan Vidar Krey
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

#include "uhub.h"

int user_manager_init(struct uhub_info* hub)
{
	assert(hub);
	hub->user_list = list_create();
	
	if (!hub->user_list)
		return -1;
		
	return 0;
}


void user_manager_shutdown(struct uhub_info* hub)
{
	assert(hub);
	
	list_clear(hub->user_list, &free);
	
	list_destroy(hub->user_list);
}


void user_manager_process(struct uhub_info* hub)
{
	assert(hub);
	
	struct linked_list* user_list = hub->user_list;
	
	time_t now = time(NULL);
	struct user* user = (struct user*) list_get_first(user_list);
	struct linked_list* tmp_list = list_create();

	if (tmp_list == NULL)
		return; /* OUCH! - We're OOM and cannot clean up! */

	while (user)
	{
		if ((user->state == state_wait_disconnect) ||
			(user->state == state_none && now - user->tm_connected > 30) ||
			(user->state != state_normal && now - user->tm_connected > 60))
		{
			uhub_log(log_info, "[%s] Disconnected", sid_to_string(user->sid));
			net_remove(user->sd);
			net_close(user->sd);
			destroy_user(user);
		}
		else
		{
			list_remove(user_list, user);
			list_append(tmp_list, user);
		}
		user = (struct user*) list_get_first(user_list);
	}

	user = list_get_first(tmp_list);
	while (user)
	{
		list_remove(tmp_list, user);
		list_append(user_list, user);
		user = list_get_first(tmp_list);
	}
	list_destroy(tmp_list);
}


struct user* create_user(struct uhub_info* hub, int sd)
{
	struct user* user = NULL;

	assert(hub);
	assert(hub->user_list);
	
	user = (struct user*) malloc(sizeof(struct user));
	

	if (user == NULL)
		return NULL; /* OOM */
	
	memset(user, 0, sizeof(struct user));
	user->sd = sd;
	user->sid = get_free_sid(hub);
	user->tm_connected = time(NULL);
	user->hub = hub;
	user->feature_cast = 0;
	
	user->send_queue = list_create();
	user->send_queue_offset = 0;
	user->send_queue_size = 0;
	
	user->recv_buf_offset = 0;
	user->recv_buf = 0;

	list_append(hub->user_list, user);

	uhub_log(log_debug, "create_user(): adding new user (sid=%d, sd=%d)", (int) user->sid, (int) user->sd);
	return user;
}


void destroy_user(struct user* user)
{
	assert(user);
	assert(user->hub);
	assert(user->hub->user_list);
	list_remove(user->hub->user_list, user);
	
	if (user->info != NULL)
	{
		send_quit_message(user->hub, user);
		command_free(user->info);
	}
	
	if (user->feature_cast)
	{
		list_clear(user->feature_cast, &free);
		list_destroy(user->feature_cast);
		user->feature_cast = 0;
	}
	
	if (user->recv_buf)
	{
		free(user->recv_buf);
		user->recv_buf = 0;
		user->recv_buf_offset = 0;
	}
	
	uhub_log(log_debug, "destroy_user(): removing user (sid=%d, sd=%d)", (int) user->sid, (int) user->sd);
	free(user);
}


struct user* get_user_by_sid(struct uhub_info* hub, sid_t sid)
{
	struct user* user = (struct user*) list_get_first(hub->user_list);
	while (user)
	{
		if (user->sid == sid)
			return user;
		user = (struct user*) list_get_next(hub->user_list);
	}
	return NULL;
}


struct user* get_user_by_nick(struct uhub_info* hub, const char* nick)
{
	struct user* user = (struct user*) list_get_first(hub->user_list);
	while (user)
	{
		if (strcmp(user->nick, nick) == 0)
			return user;
		user = (struct user*) list_get_next(hub->user_list);
	}
	return NULL;
}


struct user* get_user_by_cid(struct uhub_info* hub, const char* cid)
{
	struct user* user = (struct user*) list_get_first(hub->user_list);
	while (user)
	{
		if (strcmp(user->cid, cid) == 0)
			return user;
		user = (struct user*) list_get_next(hub->user_list);
	}
	return NULL;
}


void send_to_user(struct user* user, struct adc_command* command)
{
	int ret;
	
	command->length = strlen(command->cache);
	if (user->send_queue_size == 0)
	{
		ret = send(user->sd, command->cache, command->length, 0);
		
		if (ret == command->length)
			return;
		
		if (ret >= 0 || (ret == -1 && (errno == EAGAIN || errno == EINTR)))
		{
			command_incref(command);
			list_append(user->send_queue, command);
			user->send_queue_size += command->length - ret;
			net_modify(user->sd, evt_read | evt_close | evt_write);
			
			if (ret > 0)
				user->tm_last_write = time(NULL);
		}
		else
		{
			/* A socket error occured */
			user->state = state_wait_disconnect;
		}
	}
	else
	{
		if (user->send_queue_size + command->length > user->hub->config->max_send_buffer)
		{
			/* User is not able to swallow the data, let's disconnect. */
			user->state = state_wait_disconnect;
		}
		else
		{
			if (user->send_queue_size + command->length > user->hub->config->max_send_buffer_soft && command->priority >= 0)
			{
				/* Don't queue this message if it is low priority! */
			}
			else
			{
				command_incref(command);
				list_append(user->send_queue, command);
				user->send_queue_size += command->length;
				net_modify(user->sd, evt_read | evt_close | evt_write);
			}
		}
	}
	
}


void send_queued_data(struct user* user)
{
	struct adc_command* command;
	int ret;
	int length;
	
	command = list_get_first(user->send_queue);
	while (command)
	{
		length = command->length - user->send_queue_offset;
		ret = send(user->sd, &command->cache[user->send_queue_offset], length, 0);
		
		if (ret >= 0 || (ret == -1 && (errno == EAGAIN || errno == EINTR)))
		{
			if (ret == length)
			{
				user->send_queue_size -= ret;
				user->tm_last_write = time(NULL);
				command_free(command);
				command = 0;
				
				if (user->send_queue_size == 0)
				{
					/* Don't poll for write */
					net_modify(user->sd, evt_read | evt_close);
					break;
				}
			}
			else if (ret > 0) {
				user->send_queue_size -= ret;
				user->tm_last_write = time(NULL);
				break;
			}
		}
		else
		{
			/* A socket error occured */
			user->state = state_wait_disconnect;
			break;
		}
		command = list_get_first(user->send_queue);
	}
}


void send_to_all(struct uhub_info* hub, struct adc_command* command)
{
	assert(hub);
	assert(command);
	struct user* user = (struct user*) list_get_first(hub->user_list);
	while (user)
	{
		if (user->state == state_normal)
		{
			send_to_user(user, command);
		}
		user = (struct user*) list_get_next(hub->user_list);
	}
}


void kick_user(struct user* user)
{
	assert(user);
	
	int old_state;
	
	struct adc_command* command = adc_create_command("IQUI");
	add_argument(command, (const char*) sid_to_string(user->sid));
	
	uhub_log(log_debug, "Kicking user");
	
	old_state = user->state;
	user->state = state_wait_disconnect;
}


int user_have_feature_cast_support(struct user* user, char feature[4])
{
	char* tmp = list_get_first(user->feature_cast);
	for (; tmp; tmp = list_get_next(user->feature_cast))
	{
		if (strncmp(tmp, feature, 4) == 0)
			return 1;
	}
	
	return 0;
}



void send_feature(struct uhub_info* hub, struct adc_command* command)
{
	int do_send;
	char* tmp;
	
	struct user* user = (struct user*) list_get_first(hub->user_list);
	while (user)
	{
		if (user->state == state_normal && user->feature_cast)
		{
			do_send = 1;
			
			tmp = list_get_first(command->feature_cast_include);
			for (; tmp; tmp = list_get_next(command->feature_cast_include));
			{
				if (!user_have_feature_cast_support(user, tmp))
				{
					do_send = 0;
					break;
				}
			}
			
			if (!do_send) continue;
			
			tmp = list_get_first(command->feature_cast_exclude);
			for (; tmp; tmp = list_get_next(command->feature_cast_exclude));
			{
				if (user_have_feature_cast_support(user, tmp))
				{
					do_send = 0;
					break;
				}
			}
			
			if (!do_send) continue;
			
			send_to_user(user, command);
		}
		user = (struct user*) list_get_next(hub->user_list);
	}

}


void send_user_list(struct uhub_info* hub, struct user* target)
{
	assert(target);
	
	struct user* user = (struct user*) list_get_first(hub->user_list);
	while (user)
	{
		if (user->state == state_normal)
		{
			send_to_user(target, user->info);
		}
		user = (struct user*) list_get_next(hub->user_list);
	}
}


void send_quit_message(struct uhub_info* hub, struct user* leaving)
{
	struct adc_command* command = adc_create_command("IQUI");
	add_argument(command, (const char*) sid_to_string(leaving->sid));
	
	send_to_all(hub, command);
	command_free(command);
}


